# Generated by roxygen2: do not edit by hand

S3method(format,shinytest2_log)
S3method(print,shinytest2_log)
export(AppDriver)
export(compare_screenshot_threshold)
export(get_input_processors)
export(load_app_env)
export(load_app_support)
export(local_app_support)
export(migrate_from_shinytest)
export(platform_variant)
export(record_test)
export(register_input_processor)
export(screenshot_max_difference)
export(test_app)
export(use_shinytest2)
export(use_shinytest2_test)
export(with_app_support)
importFrom(R6,R6Class)
importFrom(callr,process)
importFrom(cli,col_blue)
importFrom(cli,col_cyan)
importFrom(cli,col_green)
importFrom(cli,col_magenta)
importFrom(cli,col_red)
importFrom(cli,col_silver)
importFrom(cli,make_ansi_style)
importFrom(lifecycle,deprecated)
importFrom(rlang,"%||%")
importFrom(rlang,":=")
importFrom(rlang,list2)
importFrom(rlang,missing_arg)
importFrom(testthat,default_reporter)
useDynLib(shinytest2, .registration = TRUE)
