## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(echo = TRUE, eval = FALSE)

## ----echo=FALSE, eval=TRUE, out.width='100%', fig.align='center', fig.alt="Screenshot of a Shiny app displaying a scatter plot with slider controls, demonstrating how exported values like plot objects and data can be tested using shinytest2"----
knitr::include_graphics("images/plot-app.png")

